/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

poll.c is for computing the chances of winning multiplayer polls,
(i.e., predicting election results,) based on favorability/approval
ratings. Favorability ratings are assumed to be a simple fractal with
Gaussian/normal distributed increments. The distribution of the
players favorability ratings in the future will be a Gaussian/normal
distribution.  If elements, Ea, and Eb, are picked, respectively, from
independent, (iid is assumed,) random processes, Ra and Rb, where Ra
and Rb have a Gaussian/normal PDF with STD Sa and Sb, and medians Ma
and Mb, with Ma >= Mb, the probability that Ea >= Eb, we are
considering Ea - Eb >= 0. The z- score of 0 is:

    z = (0 - mean) / std = (Mb - Ma) / sqrt (Sa^2 + Sb^2)

and the probability is 1 - Phi (z) where Phi (z) is the cumulative
probability distribution of the Gaussian/normal distribution.

If Sa = Sb = S:

    z = (0 - mean) / std = (Mb - Ma) / (sqrt (2) * S)

What this means is to build a table of the cumulative of the
Gaussian/normal distribution, and look up the value of the cumulative
for (Mb - Ma) / (sqrt (2) * S), which is what this program does for
each player, multiplying the chances of a player winning over all the
other players together, to get the chance of the player winning.

Note that the fractal probably has a log-normal distribution, since
negative favoribility ratings are impossible, and there is no bounds
on the upper limit, (more voters vote in the future, or those that
usually don't vote do,) and the values should be number of voters,
instead of favoribilities in percentages. Additionally, if there are
only two players, the distributions of the two players are not
independent, (since it is a zero sum game, and what one player wins,
the other loses-a negative correlation, effectively doubling the
root-mean-square of the increments of the favorability ratings.)

The tsinvest(1) program should be used for such scenarios, (that's its
forte-the "currency" would be percentage points of favorability,) but
such a precision methodology would require far more data than is
normally available in elections. This program uses the log-normal and
cumulative of the Gaussian/normal probability distribution
computations from the tsinvest program to provide an estimation of
future election winner(s) based on current favorability ratings.

The command line arguments are the favoribility ratings of
the players, and the chances of each player winning is
printed to stdout.

A note about the method of calculation of the combined probabilities
for a player:

For two players, the possibilities are:

    A > B
    B > A *

For three players, the possibilities are:

    A > B
    A > C
    B > A *
    B > C
    C > A *
    C > B *

For four players, the possibilities are:

    A > B
    A > C
    A > D
    B > A *
    B > C
    B > D
    C > A *
    C > B *
    C > D
    D > A *
    D > B *
    D > C *

In general, for n players, there are N(N - 1) rows, where N is the
number of poll values on the command line. The rows with an asterisk
are the compliment, (inverse,) of another row, (i.e., A > B and B >
A,) and the combined probability of a row and its compliment is
unity. Therefore, the sum of all rows is N(N - 1) / 2-which could be
calculated by making a pass of the data to calculate the sum of all
rows, and another pass, multiplying by the reciprocal of the sum of
all rows. Thus the program will present answers, (subject to
interpretation-for example, Player X has twice the probability of
winning as Player Y, etc.,) when the poll numbers do not total to
unity, (and the program is ratio metric so long as the "-r rms"
dimension and poll numbers on the command line are the same, i.e.,
both percent, both fractional, etc.)

For example, using fractional:

    poll -r 0.02 -t 365 0.50 0.25 0.125 0.0625 0.0625
    Player 1's chances of winning = 0.320796
    Player 2's chances of winning = 0.076919
    Player 3's chances of winning = 0.029703
    Player 4's chances of winning = 0.017288
    Player 5's chances of winning = 0.017288

Or, using percent:

    poll -r 2 -t 365 50 25 12.5 6.25 6.25
    Player 1's chances of winning = 0.320796
    Player 2's chances of winning = 0.076919
    Player 3's chances of winning = 0.029703
    Player 4's chances of winning = 0.017288
    Player 5's chances of winning = 0.017288

means the root-mean-square of the increments of the poll data is 2
percentage points, (-r 2,) a day, it is one year to the election, (-t
= 365 days,) and the first player has 50% favorability, the second
25%, and so on.

By comparison, using the log-normal distribution evolution, (which
should be used with actual voter counts, instead of
fractional; percentages will often yield adequate estimations):

    poll -l -r 2 -t 365 50 25 12.5 6.25 6.25
    Player 1's chances of winning = 0.176716
    Player 2's chances of winning = 0.099164
    Player 3's chances of winning = 0.050807
    Player 4's chances of winning = 0.023629
    Player 5's chances of winning = 0.023629

Note that the player's chances are pure probabilities. To normalize
the probabilities, (i.e., they sum to unity,) use the -N option:

    poll -N -l -r 2 -t 365 50 25 12.5 6.25 6.25
    Player 1's chances of winning = 0.472572
    Player 2's chances of winning = 0.265182
    Player 3's chances of winning = 0.135867
    Player 4's chances of winning = 0.063189
    Player 5's chances of winning = 0.063189

The normalization can be calculated, (but the poll value arguments on
the command line should sum to unity,) by noting the 50/50, (or 1 and
its complement, 0, and vice versa,) values of a player winning, P, for
the number of poll values, N, n = 2, n = 3, n = 4, ... n = N. For N =
2, P = 0.5^(2-1) = 0.5; for N = 3, P = 0.5^(3-1) = 0.25; for N = 4, P
= 0.5^(4-1) = 0.125; ...; for N = n, P = 0.5^(n-1), which would
correspond to a normalizer of 2(0.5^(2-1)) = 1 for N = 2; 3(0.5^(3-1))
= 0.75 for N = 3; 4(0.5^(4-1)) = 0.5 for N = 4; ...; n(0.5^(n-1)) for
N = n.

$Revision: 0.0 $
$Date: 2007/01/18 19:36:00 $
$Id: poll.c,v 0.0 2007/01/18 19:36:00 john Exp $
$Log: poll.c,v $
Revision 0.0  2007/01/18 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#ifdef __STDC__

#include <float.h>

#endif

#ifndef PI /* make sure PI is defined */

#define PI 3.14159265358979323846 /* pi to 20 decimal places */

#endif

#ifdef __STDC__

static char rcsid[] = "$Id: poll.c,v 0.0 2007/01/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2007, John Conover, All Rights Reserved"; /* the copyright banner */

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Compute the chances of winning a multiplayer election\n",
    "Usage: poll [-l] [-N] [-r rms] [-t time] [-V] [-v] poll1 poll2 ...\n",
    "    -l, log-normal distribution evolves over time, (normal distribution)\n",
    "    -N, normalize election probabilities to total unity\n",
    "    -r rms, root-mean-square of increments of favoribility ratings, (1)\n",
    "    -t time, time to election, (1)\n",
    "    -V, verbose output\n",
    "    -v, print the program's version information\n",
    "    poll1 poll2 ..., current poll favorability/approval ratings\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static void cumulativenormal (void); /* construct the cumulative of the normal distribution */

#else

static void print_message (); /* print any error messages */
static void cumulativenormal (); /* construct the cumulative of the normal distribution */

#endif

#define SIGMAS 7 /* 7 sigma limit, ie., -7 to +7 sigma */
#define STEPS_PER_SIGMA 100000 /* each sigma has 10000 steps of granularity */

static double cumulative[2 * SIGMAS * STEPS_PER_SIGMA]; /* the array of cumulative levels, ie., the cumulative of the normal distribution */

static int sigma_limit = 2 * SIGMAS * STEPS_PER_SIGMA; /* the array size of the array of cumulative levels, ie., 2 * SIGMAS * STEPS_PER_SIGMA, for calculation expediency */

static int cumulativeconstructed = 0; /* flag to determine whether the cumulative normal distribution array, cumulative[], has been set up, 0 = no, 1 = yes */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    int c, /* command line switch */
        i, /* counter */
        j, /* counter */
        k, /* counter */
        l = 0, /* use log-normal distribution, 0 = no, 1 = yes */
        n = 0, /* normalize probabilities, 0 = no, 1 = yes */
        v = 0, /* verbose output, 0 = no, 1 = yes */
        state, /* state counter, 0 = sum marginals, 1 = compute probabilities */
        players, /* number of players */
        retval = EARGS; /* return value, assume not enough arguments */

    double r = (double) 1.0, /* root-mean-square of favoribility ratings */
           t = (double) 1.0, /* time to election */
           sigma = (double) -(2 * SIGMAS), /* the sigma value to compute the cumulative of the normal distribution */
           marginal, /* marginal probability of player winning against a specific player */
           normalizer, /* normalization constant = sum of probabilites for a player to win election */
           favorability[BUFSIZ * 2], /* favorability of players */
           P = (double) 1.0, /* probability of player winning */
           sigmas = (double) 1.0; /* standard deviations */

    while ((c = getopt (argc, argv, "Nlr:t:Vv")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'l': /* request for log-normal distribution? */

                l = 1; /* yes, set use log-normal distribution, 0 = no, 1 = yes */
                break;

            case 'N': /* request to normalize probabilities? */

                n = 1; /* yes, set normalize probabilities, 0 = no, 1 = yes */
                break;

            case 'r': /* request for root-mean-square of favoribility ratings? */

                r = atof (optarg); /* yes, save the root-mean-square of favoribility ratings */
                break;

            case 't': /* request for time to election? */

                t = atof (optarg); /* yes, save the time to election */
                break;

            case 'V': /* request for verbose output? */

                v = 1; /* yes, set verbose output, 0 = no, 1 = yes */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if ((sigma = sqrt ((double) 2.0) * r * sqrt ((double) t)) > (double) 0) /* sigma valid? */
    {

        if (l == 1) /* log normal distribution? */
        {
            sigma = log (sigma); /* take the log of the sigma */
        }

        if ((players = (argc - optind)) > 0) /* enough players, (3 or more)? */
        {

            if (players >= 1) /* one player has 100% chance of winning */
            {
                retval = NOERROR; /* assume no error */

                for (i = 0; i < players; i ++) /* for each player */
                {
                    favorability[i] = favorability[i + players] = atof (argv[optind ++]); /* save the player's favorability rating, twice in order */
                }

                if (cumulativeconstructed == 0) /* cumulative normal distribution array, cumulative[], been set up? 0 = no, 1 = yes */
                {
                    cumulativenormal (); /* no, set up the cumulative normal distribution array, cumulative[] */
                }

                normalizer = (double) 0.0; /* normalization constant = sum of probabilites for a player to win election */

                for (state = 0; state <= 1; state ++) /* for each state; state counter, 0 = sum marginals, 1 = compute probabilities */
                {

                    for (j = 0; j < players; j ++) /* for each player */
                    {
                        k = players + j; /* k = last player, j = first player */
                        P = (double) 1.0; /* probability of first player winning */

                        for (i = j + 1; i < k; i ++) /* for each of the other players */
                        {

                            if (l == 1) /* log normal distribution? */
                            {
                                sigmas = (log (favorability[i]) - log (favorability[j])) / sigma; /* subtract the logs of the favorabilities, and divided by the sigma */
                            }
                            else
                            {
                                sigmas = (favorability[i] - favorability[j]) / sigma; /* subtract the favoribilites, and divide by the sigma */
                            }

                            if (sigmas < (double) -SIGMAS) /* less than -SIGMAS? */
                            {
                                marginal = (double) 1.0; /* marginal probability = 0 */
                            }

                            else if (sigmas >= (double) SIGMAS) /* past SIGMAS? */
                            {
                                marginal = (double) 0.0; /* probability 1 */
                            }

                            else
                            {
                                marginal = ((double) 1.0 - cumulative[(int) floor (STEPS_PER_SIGMA * (SIGMAS + sigmas))]); /* cumulative of the normal distribution for this sigma */
                            }

                            P = P * marginal; /* probability of player winning against all others*/

                            if ((v == 1) && (state == 1)) /* verbose output, 0 = no, 1 = yes? */
                            {
                                (void) printf ("Player %d's chances of winning against player %d = %f, (partial for player %d = %f, sigma = %f, sigmas = %f, normalizer = %f)\n", j + 1, i < players ? i + 1: i - players + 1, marginal, j + 1, P, sigma, sigmas, normalizer); /* print it */
                            }

                        }

                        if (state == 0) /* state counter, 0 = sum marginals, 1 = compute probabilities */
                        {
                            normalizer = normalizer + P; /* normalization constant = sum of probabilites for a player to win election */
                        }

                        else
                        {

                            if (n == 0) /* normalize probabilities, 0 = no, 1 = yes? */
                            {
                                (void) printf ("Player %d's chances of winning = %f\n", j + 1, P);  /* print it */
                            }

                            else
                            {
                                (void) printf ("Player %d's chances of winning = %f\n", j + 1, P / normalizer);  /* print it */
                            }

                        }

                    }

                }

            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Construct the cumulative of the normal distribution.

static void cumulativenormal (void);

I) The process chosen uses a binary search of the cumulative of the
normal distribution. First a "graph" or "table" (ie., like the table
above,) is constructed in memory of the cumulative of the normal
distribution. This "graph" is then searched to compute the value of
the cumulative of the normal distribution.

II) Constructs the array, cumulative[], which is used as a lookup
mechanism for the cumulative of a normal distribution.

III) This is simply a lookup table, with the implicit index
representing a sigma value.  Each sigma is divided into
STEPS_PER_SIGMA many steps.

Returns nothing.

*/

#ifdef __STDC__

static void cumulativenormal (void)

#else

static void cumulativenormal ()

#endif

{
    int i; /* step counter through cumulative of normal distribution */

    double scale = (double) 1.0 / sqrt ((double) 2.0 * (double) PI), /* scaling factor for cumulative of normal distribution, for math expediency */
           del = (double) 1.0 / (double) STEPS_PER_SIGMA, /* increment of granularity of x axis of cumulative of normal distribution */
           x = (double) -SIGMAS, /* x variable in cumulative of normal distribution */
           cumulativesum = (double) 0.0; /* cumulative of normal distribution, begins at -7 sigma of the distribution, ie., x = -7 */

    for (i = 0; i < sigma_limit; i ++) /* for each step in the cumulative of the normal distribution */
    {
        cumulativesum = cumulativesum + ((scale * (exp ((- (x * x)) / ((double) 2.0)))) / (double) STEPS_PER_SIGMA); /* add the value of the normal distribution for this x to the cumulative of the normal distribution */
        cumulative[i] = cumulativesum; /* save the value of the cumulative of the normal distribution for this x */
        x = x + del; /* next increment of x in the cumulative of the normal distribution */
    }

    cumulativeconstructed = 1; /* set the flag to determine whether the cumulative normal distribution array, cumulative[], has been set up, 0 = no, 1 = yes */
}
